/*
 * Copyright (c) 2004-2005 Sergey Lyubka <valenok@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*
 * Simple and portable HTTP server
 * Features:  CGI, SSL, Basic Auth, Cookies
 *
 * Compilation:
 *    SSL support:	-DWITH_SSL -lssl -lcrypto
 *    windows GUI:	-DWITH_GUI
 *    No CGI:		-DNO_CGI
 *    No Basic auth:	-DNO_AUTH
 *
 * Changelog:
 *   1.1	Multiple index files support
 *   1.2	Redirecting errors to a file.
 *		Fixed 'page contains no data' bug.
 *   1.3	Better error reporting (404, 500)
 *		Fixed CGI handling for win32, when CGI file is a native
 *		executable, not a script
 *   1.4	All replies are now set to HTTP/1.1 instead of HTTP/1.0
 *		For regular files, added headers: Content-Length, Date,
 *		Last-Modified, Etag.
 *		Fixed win32 CGI: passing environment to spawaned script.
 *   1.5	After start, change current dir to docroot
 *		Optimized environment build for CGI. Added few more
 *		env variables (REMOTE_ADDR, USER_AGENT, PATH_INFO, etc)
 *		Access log file opened once at start time, not on every request
 *		Added GUI for Windows
 *   1.6	Fixed CGI headers output (newline was missed)
 *		Thanks to Arno Schuring and Frash.
 *   1.7	Directory Listing support.
 *		'Kernel' rewritten. there is no global select() anymore.
 *		win32 GUI settings are stored in HKCU, thanks Ben Truscott
 *   1.8	Basic authentication is gone. Digest auth (MD5) is used instead.
 *   1.9	Support for user-defined mime types file
 */

#define	VERSION		"1.9"			/* Version */
#define	ENVSIZ		(16 * 1024)		/* Environment size */
#define	IOBUFSIZ	(16 * 1024)		/* IO buffer size */
#define	REMUSERSIZ	64			/* Remote user name maxsize */
#define	PATHSIZ		1024			/* PATH_MAX, max path size */
#define	MAXHDRS		64			/* Maximum number of headers */
#define	NELEMS(ar)	(sizeof(ar) / sizeof(ar[0]))

#ifdef _WIN32					/* Windows specific */
#include <winsock.h>
#include <windows.h>
#include <process.h>
#include <io.h>
#define	ERRNO				GetLastError()
typedef int				socklen_t;
typedef unsigned int			uint32_t;
typedef unsigned short			uint16_t;
#define	DIRSEP				'\\'
#define	waitpid(a,b,c)			0
#define	EWOULDBLOCK			WSAEWOULDBLOCK
#define	SIGCHLD				0
#define	SIGPIPE				SIGTERM
#define	snprintf			_snprintf
#define	vsnprintf			_vsnprintf
#define	setuid(a)			0
#define	usleep(x)			Sleep((x) / 1000)

/* POSIX dirent interface */
struct dirent {
	char	*d_name;
};

typedef struct DIR {
	long			handle;
	struct _finddata_t	info;
	struct dirent		result;
	char			*name;
} DIR;

#else						/* UNIX specific */

#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <unistd.h>
#include <dirent.h>
#define	O_BINARY			0
#define	closesocket(a)			close(a)
#define	DIRSEP				'/'
#define	ERRNO				errno
#define	PULL(a,b,c)			read(a, b, c)
#define	PUSH(a,b,c)			write(a, b, c)
#endif	/* _WIN32 */

#include <sys/types.h>				/* Common */
#include <sys/stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <assert.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <fcntl.h>

#ifdef WITH_DMALLOC
#include <dmalloc.h>
#endif /* WITH_DMALLOC */

#ifdef WITH_SSL
#include <openssl/ssl.h>
#else						/* Stubs for non-SSL build */
typedef void *SSL_CTX;
typedef void *SSL;
#define	SSL_free(x)				x = x
#define	SSL_read(a,b,c)				(a = a, b = b, c = c)
#define	SSL_write(a,b,c)			(a = a, b = b, c = c)
#define	SSL_accept(a)				0
#define	SSL_load_error_strings()		(void) 0
#define	SSL_set_fd(a,b)				0
#define	SSL_get_fd(a)				-1
#define	SSLeay_add_ssl_algorithms()		(void) 0
#define	SSL_CTX_use_certificate_file(a,b,c)	0
#define	SSL_CTX_use_PrivateKey_file(a,b,c)	0
#define	SSL_CTX_new(a)				0
#define	SSL_FILETYPE_PEM			0
#define	SSL_new(a)				0
#define	SSL_ERROR_WANT_WRITE			-1
#define	SSL_ERROR_WANT_READ			-1
#define	SSL_get_error(a,b)			0
#define	SSL_shutdown(a)				(void) 0
#endif /* WITH_SSL */

/*
 * Macros for doubly-linked circular linked lists.
 */
struct llhead {
	struct llhead *prev, *next;
};

#define	LL_INIT(N)	((N)->next = (N)->prev = (N))

#define LL_HEAD(H)	struct llhead H = { &H, &H }

#define	LL_ENTRY(P,T,N)	((T *)((char *)(P) -(unsigned long)(&((T *) 0)->N)))

#define	LL_TAIL(H, N)							\
	do {								\
		((H)->prev)->next = (N);				\
		(N)->prev = ((H)->prev);				\
		(N)->next = (H);					\
		(H)->prev = (N);					\
	} while (0)

#define	LL_DEL(N)							\
	do {								\
		((N)->next)->prev = ((N)->prev);			\
		((N)->prev)->next = ((N)->next);			\
		LL_INIT(N);						\
	} while (0)

#define	LL_FOREACH(H,N)	for (N = (H)->next; N != (H); N = (N)->next)

#define LL_FOREACH_SAFE(H,N,T)						\
	for (N = (H)->next, T = (N)->next; N != (H);			\
			N = (T), T = (N)->next)

/* General vector */
struct vec {
	size_t		len;
	const void	*ptr;
};

/* HTTP URL */
struct url {
	struct vec	proto;
	struct vec	user;
	struct vec	pass;
	struct vec	host;
	struct vec	port;
	struct vec	uri;
};

/* HTTP header */
struct hthdr {
	struct vec	name;
	struct vec	value;
};

/* HTTP information */
struct hti {
	struct vec	method;
	struct vec	url;
	struct vec	proto;
	struct hthdr	hh[MAXHDRS];	/* Headers */
	int		nhdrs;		/* Actual number of headers */
	int		reqlen;		/* Request line length */
	int		totlen;		/* request len + headers len */
};

/* HTTP header callback */
typedef void (*hcb_t)(const struct hthdr *header, void *userdata);


/* Unified socket address */
struct sa {
	socklen_t	len;
	union {
		struct sockaddr		sa;
		struct sockaddr_in	sin;
#ifdef WITH_IPV6
		struct sockaddr_in6	sin6;
#endif /* WITH_IPV6 */
	} u;
};

/*
 * Our abstraction is as follows:
 * Two communicating peers are represented as 'local' and 'remote' IO buffers.
 * IO buffer consists two channels (read and write), and character buffer.
 * Read channel reads into an IO buffer. Write channel gets data from
 * elsewhere (peer's IO buffer) and writes it to peer.
 * For example, when regular file was requested, one peer is the opened
 * regular file, and the other is opened socket to the remote host.
 */

/* One-way channel, reader or writer */
struct channel {
	int	(*iofunc)(void *,char*,size_t);	/* Read or write function */
	void	(*cleanup)(void *);		/* Cleanup function */
	void	*private;			/* Private data */
};

/* I/O buffer */
struct iobuf {
	struct channel	read;			/* Read channel */
	struct channel	write;			/* Write channel */
	int		shared;			/* Read and write channels */
						/* share same private data */
	int		done;			/* I/O finished */
	char		buf[IOBUFSIZ];		/* Buffer */
	size_t		nread;			/* Bytes read */
	size_t		nwritten;		/* Bytes written */
};

/* Connection descriptor */
struct conn {
	struct llhead	link;			/* Connections chain */
	struct sa	sa;			/* Remote socket address */
	struct iobuf	local;			/* Local end io descriptor */
	struct iobuf	remote;			/* Remote connection */
	struct hti	hti;			/* Parsed request */
	struct vec	uri;			/* URI vector */
	struct vec	req;			/* stuff after ? */
	struct vec	auth;			/* Auth vector */
	struct vec	cookie;			/* Cookie: */
	struct vec	clength;		/* Content-Length: */
	struct vec	ctype;			/* Content-Type:  */
	struct vec	status;			/* Status: */
	struct vec	location;		/* Location: */
	struct vec	useragent;		/* User-Agent: */
	struct vec	referer;		/* Referer: */
	time_t		ims;			/* If-Modified-Since: */
	time_t		expire;			/* Expiration time */
	char		user[REMUSERSIZ];	/* Authorized user */
	int		nexpected;		/* headers + content length */

	/* Flags */
	char		gotrequest;		/* Request parsed flag */
	char		gotreply;		/* Got CGI reply headers */
	char		cgimode;		/* CGI */
	char		sslaccepted;		/* SSL_accept() succeeded */
};

/* Data to be passed to directory_read etc */
struct dir_data {
	DIR	*dirp;
	char	*path;
};

/* Mime type entry */
struct mimetype {
	struct mimetype	*next;
	char		*ext;			/* File extention */
	char		*mime;			/* Mime type */
	size_t		extlen;			/* Extention length */
};

/*
 * Globals
 */
static const char	*docroot = ".";		/* Document root */
static const char	*certfile ="shttpd.pem";/* Certificate file */
static const char	*htpass = ".htpasswd";	/* per-directory .htpasswd */
static const char	*globhtpass;		/* Global .htpasswd file */
static const char	*domain = "auth";	/* Server name (auth domain) */
static const char	*ifile =
	"index.html,index.cgi,index.php";	/* Index file */
static const char	*logfile = "shttpd.log";/* Access Log file */
static const char	*errfile;		/* Error Log file */
static FILE		*logfp;			/* Log stream */
static const char	*cgipat = ".cgi";	/* CGI file pattern */
static uint16_t		lport = 80;		/* Listening port */
static SSL_CTX		*ctx;			/* SSL context */
static int		uid;			/* Runtime User ID  */
static int		quit;			/* Exit flag */
static int		dossl;			/* SSL mode */
static int		do_dir;			/* Do directory listing */
static int		ttl = 300;		/* Max idle time */
static int		active;			/* Data is transferring */
static char		cwd[PATHSIZ];		/* Current directory */
static time_t		now;			/* Current time */
static const char	*mimefile;		/* Mime types file */
static struct mimetype	*mimetypes;		/* Known mime types */
static LL_HEAD(conns);				/* List of connections */

/*
 * Case-insensitive string comparison, a-la strcmp()
 */
static int
casecmp(register const char *s1, register const char *s2)
{
	for (; *s1 != '\0' && *s2 != '\0'; s1++, s2++)
		if (tolower(*s1) != tolower(*s2))
			break;
	
	return (*s1 - *s2);
}

/* Case insensitive memory comparison */
static int
ncasecmp(register const char *s1, register const char *s2, size_t len)
{
	register const char	*e = s1 + len - 1;
	
	for (; s1 < e && tolower(*s1) == tolower(*s2); s1++, s2++) ;

	return (tolower(*s1) - tolower(*s2));
}

/* Compare two vectors case-insensitive */
static int
vcasecmp(const struct vec *v1, const struct vec *v2)
{
	return (v1->len == v2->len && ncasecmp(v1->ptr, v2->ptr, v1->len) == 0);
}

/* Return integer represented by vector */
static int
vtoint(const struct vec *v)
{
	int	i, n;

	for (i = n = 0; i < (int) v->len; i++) {
		n *= 10;
		n += ((char *) v->ptr)[i] - '0';
	}
	
	return (n);
}

/* Return ptr to character `ch' in a vector, or NULL */
static const char *
vchr(const struct vec *v, char ch)
{
	register const char *s, *e, *p;

	for (p = NULL, s = v->ptr, e = s + v->len; s < e; s++)
		if (*s == ch) {
			p = s;
			break;
		}

	return (p);
}

/*
 * Assume the given string `s' of length `len' is pointing to valid URL,
 * parse this URL, filling the passed struct url u
 * Return -1 if the URL is invalid, or the URL length.
 */
static int
hturl(const char *s, int len, struct url *u)
{
	register const char	*p, *e;
	struct vec		*v, nil = { 0, 0 };
	
	(void) memset(u, 0, sizeof(*u));

	/* Now, dispatch URI */
	for (p = s, e = s + len, v = &u->proto; p < e; p++) {
		switch (*p) {
		
		case ':':
			if (v == &u->proto) {
				if (&p[2] < e && p[1] == '/' && p[2] == '/') {
					p += 2;
					v = &u->user;
				} else {
					u->user = u->proto;
					u->proto = nil;
					v = &u->pass;
				}
			} else if (v == &u->user) {
				v = &u->pass;
			} else if (v == &u->host) {
				v = &u->port;
			} else if (v == &u->uri) {
				/* : is allowed in path or query */
				v->len++;
			} else {
				return (-1);
			}
			break;
		
		case '@':
			if (v == &u->proto) {
				u->user = u->proto;
				u->proto = nil;
				v = &u->host;
			} else if (v == &u->pass || v == &u->user) {
				v = &u->host;
			} else if (v == &u->uri) {
				/* @ is allowed in path or query */
				v->len++;
			} else {
				return (-1);
			}
			break;
		
		case '/':
#define	SETURI()	v = &u->uri; v->ptr = p; v->len = 1
			if ((v == &u->proto && u->proto.len == 0) ||
			    v == &u->host || v == &u->port) {
				SETURI();
			} else if (v == &u->user) {
				u->host = u->user;
				u->user = nil;
				SETURI();
			} else if (v == &u->pass) {
				u->host = u->user;
				u->port = u->pass;
				u->user = u->pass = nil;
				SETURI();
			} else if (v == &u->uri) {
				/* / is allowed in path or query */
				v->len++;
			} else {
				return (-1);
			}
			break;
		
		default:
			if (!v->ptr)
				v->ptr = p;
			v->len++;
		}
	}

	if (v == &u->proto && v->len > 0) {
		v = ((char *) v->ptr)[0] == '/' ? &u->uri : &u->host;
		*v = u->proto;
		u->proto = nil;
	} else if (v == &u->user) {
		u->host = u->user;
		u->user = nil;
	} else if (v == &u->pass) {
		u->host = u->user;
		u->port = u->pass;
		u->user = u->pass = nil;
	}

	return (p - s);
}

/*
 * Parse given HTTP request.
 * Fill struct hti during parse.
 * For every header, call `cb'.
 * Return number of parsed bytes (normally, full request length)
 */
static int
htparse(const char *s, int len, struct hti *info, hcb_t cb, void *data)
{
	register const char	*p, *e = s + len;
	struct vec		h, v, vec[3] = {{0,0}, {0,0}, {0,0}};
	struct hthdr		*hdr;
	int			i;
	enum { WAIT, CONT, HDR, HDRVAL } state;

	(void) memset(info, 0, sizeof(*info));

	for (state = WAIT, i = 0, p = s; p < e; p++)
		switch (state) {
		
		case WAIT:
			if (*p != ' ') {
				state = CONT;
				vec[i].ptr = p;
				vec[i].len = 0;
			}
			break;
		
		case CONT:
			vec[i].len++;
			if (*p == ' ' && i < 2) {
				state = WAIT;
				i++;
			} else if (*p == '\n') {
				/* End request */
				info->reqlen = p - s + 1;
				if (vec[i].len > 0 && p[-1] == '\r')
					vec[i].len--;
				h.ptr = p + 1;
				h.len = v.len = 0;
				v.ptr = NULL;
				info->method = vec[0];
				info->url = vec[1];
				info->proto = vec[2];
				state = HDR;
			}
			break;
		
		case HDR:
			if (*p == ':') {
				v.ptr = p + 1;
				state = HDRVAL;
			} else if (*p == '\n') {
				/* End headers */
				info->totlen = p - s + 1;
				return (info->totlen);
			} else if (v.len == 0 && isspace(*p)) {
				/* skip initial whitespaces */
				h.ptr = p + 1;
			} else {
				h.len++;
			}
			break;
		
		case HDRVAL:
			if (*p == '\n') {
				if (v.len && p[-1] == '\r')
					v.len--;
				/* Save header */
				hdr = &info->hh[info->nhdrs++];
				if (info->nhdrs < (int)
				    (sizeof(info->hh) / sizeof(info->hh[0]))) {
					hdr->name = h;
					hdr->value = v;
					if (cb)
						cb(hdr, data);
				}
				h.ptr = p + 1;
				v.ptr = NULL;
				h.len = v.len = 0;
				state = HDR;
			} else if (v.len == 0 && isspace(*p)) {
				/* skip initial whitespaces */
				v.ptr = p + 1;
			} else {
				v.len++;
			}
			break;
		}

	return (-1);
}

/*
 * SIGTERM, SIGINT signal handler
 */
static void
sigterm(int signo)
{
	quit = signo;
}

/*
 * Grim reaper of innocent children: SIGCHLD signal handler
 */
static void
sigchild(int signo)
{
	while (waitpid(-1, &signo, WNOHANG) > 0) ;
}

/*
 * Log function
 */
static void
elog(int fatal, const char *fmt, ...)
{
	char	msg[IOBUFSIZ];
	int	n;
	va_list	ap;

#if 1
	if (!fatal)		/* Do not show debug messages */
		return;
#endif

	n = snprintf(msg, sizeof(msg), "%lu ", (unsigned long) now);

	va_start(ap, fmt);
	(void) vsnprintf(msg + n, sizeof(msg) - n, fmt, ap);
	va_end(ap);

#if defined(_WIN32) && defined(WITH_GUI)
	if (fatal)
		MessageBox(NULL, msg, "Error", MB_OK | MB_ICONEXCLAMATION);
#endif /* _WIN32 && WITH_GUI */
	
	(void) fprintf(stderr, "%s\n", msg);
	(void) fflush(stderr);

	if (fatal)
		exit(EXIT_FAILURE);
}

static void
usage(const char *prog)
{
	(void) fprintf(stderr,
	    "shttpd version %s\n"
	    "usage: %s [OPTIONS]\n"
	    "-d directory	document root (dflt: %s)\n"
	    "-p port		listen on port (dflt: 80 or 443)\n"
	    "-l logfile 	log file (dflt: %s)\n"
	    "-e elogfile 	errorlog file (dflt: none)\n"
	    "-i file1,.. 	index file (dflt: %s)\n"
#ifndef NO_CGI
	    "-c cgipattern	CGI file pattern (dflt: %s)\n"
#endif /* NO_CGI */
#ifndef NO_AUTH
	    "-P passfile	authorization passwords file (dflt: none)\n"
#endif /* NO_AUTH */
	    "-u numeric_uid	run-time UID (dflt: none)\n"
	    "-D 		do directory listing (dflt: no)\n"
	    "-N servername 	server domain name (auth realm) (dflt: %s)\n"
	    "-m mime-file 	mime types file (dflt: none)\n"
#ifdef WITH_SSL
	    "-s certfile	SSL certfileificate (dflt: %s)\n"
	    "-S 		use SSL (dflt: NO)\n"
#endif /* WITH_SSL */
	    , VERSION, prog, docroot, logfile, ifile
#ifndef NO_CGI
	    , cgipat
#endif /* NO_CGI */
	    , domain
#ifdef WITH_SSL
	    , certfile
#endif /* WITH_SSL */
	    );

	exit(EXIT_FAILURE);
}

/*
 * Write an HTTP access log into a file `logfile'
 */
static void
accesslog(const struct conn *c, int status)
{
	char	date[64];

	if (logfp != NULL) {
		(void) strftime(date, sizeof(date),
		    "%d/%b/%Y %H:%M:%S", localtime(&now));
		(void) fprintf(logfp, "%.15s - %s [%s] \"%.*s %.*s %.*s\" %d\n",
		    inet_ntoa(c->sa.u.sin.sin_addr),
		    c->user[0] == '\0' ? "-" : c->user,
		    date,
		    c->hti.method.len, (char *) c->hti.method.ptr,
		    c->hti.url.len, (char *) c->hti.url.ptr,
		    c->hti.proto.len, (char *) c->hti.proto.ptr, status);
		(void) fflush(logfp);
	}
}

/*
 * Put given file descriptor in blocking (block == 1)
 * or non-blocking (block == 0) mode.
 * Return 0 if success, or -1
 */
static int
nonblock(int fd)
{
#ifdef	_WIN32
	unsigned long	on = 1;

	return (ioctlsocket(fd, FIONBIO, &on));
#else
	int	flags, retval = 0;

	if ((flags = fcntl(fd, F_GETFL, 0)) == -1) {
		elog(0, "nonblock: fcntl(%d, F_GETFL): %s", fd,strerror(ERRNO));
		retval--;
	} else {
		flags |= O_NONBLOCK;
		
		/* Apply the mode */
		if (fcntl(fd, F_SETFL, flags) != 0) {
			elog(0, "nonblock: fcntl(%d, F_SETFL): %s",
			    fd, strerror(ERRNO));
			retval--;
		}
	}

	return (retval);
#endif /* _WIN32 */
}

#ifdef _WIN32

#ifndef __GNUC__
static int
chdir(const char *dir)
{
	return (SetCurrentDirectory(dir) == 0);
}

static char *
getcwd(char *dst, size_t dstlen)
{
	return (GetCurrentDirectory(dstlen, dst) == 0 ? NULL : dst);
}
#endif /* __GNUC__ */

static DIR *
opendir(const char *name)
{
	DIR		*dir = NULL;
	size_t		base_length;
	const char	*all;

	if (name && name[0]) {
		base_length = strlen(name);
		all = strchr("/\\", name[base_length - 1]) ? "*" : "/*";

		if ((dir = malloc(sizeof *dir)) != NULL &&
		    (dir->name = malloc(base_length + strlen(all) + 1)) != 0) {
			(void) strcat(strcpy(dir->name, name), all);

			if ((dir->handle = (long) _findfirst(dir->name,
			    &dir->info)) != -1) {
				dir->result.d_name = 0;
			} else {
				free(dir->name);
				free(dir);
				dir = 0;
			}
		} else {
			free(dir);
			dir = NULL;
			errno = ENOMEM;
		}
	} else {
		errno = EINVAL;
	}

	return (dir);
}

static int
closedir(DIR *dir)
{
	int result = -1;

	if (dir) {
		if(dir->handle != -1)
			result = _findclose(dir->handle);

		free(dir->name);
		free(dir);
	}

	if (result == -1) 
		errno = EBADF;

	return (result);
}

static struct dirent *
readdir(DIR *dir)
{
	struct dirent *result = 0;

	if (dir && dir->handle != -1) {
		if(!dir->result.d_name ||
		    _findnext(dir->handle, &dir->info) != -1) {
			result = &dir->result;
			result->d_name = dir->info.name;
		}
	} else {
		errno = EBADF;
	}

	return (result);
}

#ifdef WITH_GUI

/*
 * Dialog box control IDs
 */
#define	ID_GROUP		100
#define	ID_BTN_QUIT		101
#define	ID_BTN_RUN		102
#define	ID_STATIC		103
#define	ID_EDIT_PORT		104
#define	ID_EDIT_WWWROOT		105
#define	ID_EDIT_LOGFILE		106
#define	ID_EDIT_INDEX		107
#define	ID_EDIT_CGIPAT		108
#define	ID_STATUSBAR		109
#define	ID_EDIT_LOG		110
#define	ID_EDIT_ELOG		111
#define	ID_EDIT_CERT		112
#define	ID_CB_DOSSL		113
#define	ID_CB_DIRLIST		114
	
static const char	*regpath = "Software\\shttpd";

/*
 * retrieve the registry value for the control with identifier `id'
 */
void
reggetval(HWND hDlg, int id, const char *dflt)
{
	static char		value[1024];
	char			name[64];
	HKEY			key;
	DWORD			type, len;

	value[0] = '\0';
	len = sizeof(value);
	(void) snprintf(name, sizeof(name), "%d", id);
	
	/* try to open path */
	if (RegCreateKeyEx(HKEY_CURRENT_USER, regpath, 0L, NULL, 0L,
	    KEY_ALL_ACCESS, NULL, &key, NULL) == ERROR_SUCCESS) {
		RegQueryValueEx(key, name, NULL, &type, (BYTE *) value, &len);
		RegCloseKey(key);
	}
	
	if (value[0] == '\0')
		(void) strncat(value, dflt, sizeof(value));
	
	if (id == ID_CB_DOSSL || id == ID_CB_DIRLIST)
		CheckDlgButton(hDlg, id,
		    value[0] == '1' ? BST_CHECKED : BST_UNCHECKED);
	else
		SetDlgItemText(hDlg, id, value);
}

/*
 * Save individual control's text into registry
 */
void
setregval(HWND hDlg, int id, char *text, size_t textlen)
{
	char	name[64];
	DWORD	len;
	HKEY	key;

	if (id != ID_CB_DOSSL && id != ID_CB_DIRLIST)
		GetDlgItemText(hDlg, id, text, textlen);
	(void) snprintf(name, sizeof(name), "%d", id);
	
	if (RegCreateKeyEx(HKEY_CURRENT_USER, regpath, 0L, NULL, 0L,
	    KEY_ALL_ACCESS, NULL, &key, NULL) == ERROR_SUCCESS) {
		len = (DWORD) strlen(text);
		RegSetValueEx(key, name, 0L, REG_SZ, (BYTE *) text, len);
		RegCloseKey(key);
	}
}

/*
 * The dialog box procedure.
 */
BOOL CALLBACK
DlgProc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
#define	TEXTL	128
	static char	port[TEXTL], root[TEXTL], index[TEXTL],
			pat[TEXTL], log[TEXTL], errlog[TEXTL], cert[TEXTL];
	static int	running, tid;
	char		title[256];
	extern int	main();

	switch (msg) {
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case ID_BTN_QUIT:
			PostQuitMessage(0);
			break;
		case ID_CB_DOSSL:
			dossl = IsDlgButtonChecked(hDlg, ID_CB_DOSSL);
			SetDlgItemText(hDlg, ID_EDIT_PORT, dossl ? "443": "80");
			EnableWindow(GetDlgItem(hDlg,ID_EDIT_CERT), dossl);
			break;
		case ID_CB_DIRLIST:
			do_dir = IsDlgButtonChecked(hDlg, ID_CB_DIRLIST);
			break;
		case ID_BTN_RUN:
			EnableWindow(GetDlgItem(hDlg,ID_EDIT_PORT), running);
			EnableWindow(GetDlgItem(hDlg,ID_EDIT_WWWROOT), running);
			EnableWindow(GetDlgItem(hDlg,ID_EDIT_CGIPAT), running);
			EnableWindow(GetDlgItem(hDlg,ID_EDIT_INDEX),running);
			EnableWindow(GetDlgItem(hDlg,ID_BTN_QUIT),running);
			EnableWindow(GetDlgItem(hDlg,ID_EDIT_LOG),running);
			EnableWindow(GetDlgItem(hDlg,ID_EDIT_ELOG),running);
			EnableWindow(GetDlgItem(hDlg,ID_EDIT_CERT),running);
			EnableWindow(GetDlgItem(hDlg,ID_CB_DOSSL),running);
			EnableWindow(GetDlgItem(hDlg,ID_CB_DIRLIST),running);
			SetDlgItemText(hDlg, ID_BTN_RUN,
			    running ? "Run" : "Quit");

			if (running) {
				PostQuitMessage(0);
				running = 0;
				quit = 1;
			} else {
				extern void doit(void);
				running = 1;
				quit = 0;
				setregval(hDlg, ID_EDIT_PORT, port, TEXTL);
				setregval(hDlg, ID_EDIT_WWWROOT, root, TEXTL);
				setregval(hDlg, ID_EDIT_INDEX, index, TEXTL);
				setregval(hDlg, ID_EDIT_CGIPAT, pat, TEXTL);
				setregval(hDlg, ID_EDIT_LOG, log, TEXTL);
				setregval(hDlg, ID_EDIT_ELOG, errlog, TEXTL);
				setregval(hDlg, ID_EDIT_CERT, cert, TEXTL);
				setregval(hDlg, ID_CB_DOSSL, dossl?"1":"0", 1);
				setregval(hDlg, ID_CB_DIRLIST,do_dir?"1":"0",1);
				
				/* Setup global according to the GUI */
				docroot = root;
				lport = (uint16_t) atoi(port);
				ifile = index;
				cgipat = pat;
				logfile = log;
				certfile = cert;
				if (errlog[0] != '\0')
					errfile = errlog;
				CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)
				    doit, NULL, 0, &tid);
			}
			break;
		}
		break;
	case WM_INITDIALOG:
		
		/* Set window title */
		(void) snprintf(title, sizeof(title),
		    "shttpd v.%s web server (%s SSL)", VERSION,
#ifdef WITH_SSL
		    "with"
#else
		    "without"
#endif /* WITH_SSL */
		    );
		SetWindowText(hDlg, title);
		SetFocus(GetDlgItem(hDlg, ID_BTN_RUN));
		
		/* Get default values from the registry */
		reggetval(hDlg, ID_EDIT_WWWROOT, docroot);
		reggetval(hDlg, ID_EDIT_PORT, "80");
		reggetval(hDlg, ID_EDIT_INDEX, ifile);
		reggetval(hDlg, ID_EDIT_CGIPAT, cgipat);
		reggetval(hDlg, ID_EDIT_LOG, logfile);
		reggetval(hDlg, ID_EDIT_ELOG, errlog);
		reggetval(hDlg, ID_EDIT_CERT, certfile);
		reggetval(hDlg, ID_CB_DOSSL, "0");
		reggetval(hDlg, ID_CB_DIRLIST, "0");

		EnableWindow(GetDlgItem(hDlg,ID_EDIT_CERT),
		    IsDlgButtonChecked(hDlg, ID_CB_DOSSL));
#ifndef WITH_SSL
		EnableWindow(GetDlgItem(hDlg,ID_CB_DOSSL), 0);
		EnableWindow(GetDlgItem(hDlg,ID_EDIT_CERT), 0);
#endif /*!WITH_SSL */

		break;
	case WM_QUIT:
		DestroyWindow(hDlg);
		break;
	case WM_CLOSE:
		PostQuitMessage(0);
		break;
	default:
		break;
	}

	return (FALSE);
}

int WINAPI
WinMain(HINSTANCE inst, HINSTANCE prev, LPSTR cmdline, int show)
{
	struct dlgitem {
		DLGITEMTEMPLATE	item;
		WORD		signature, class;
		wchar_t		title[12];
		WORD		data;
	};
	struct dlgheader {
		DLGTEMPLATE	templ;		/* 18 bytes */
		WORD		menu, class;
		wchar_t		caption[1];
		WORD		fontsiz;
		wchar_t		fontface[7];
	};

	/* Create dialog box template in memory */
	struct {
		struct dlgheader	header;
		struct dlgitem		items[20];
	} dialog = {
		{{ WS_CAPTION | WS_POPUP | WS_SYSMENU | WS_VISIBLE | DS_SETFONT
		    | WS_MINIMIZEBOX, 0, 20, 200, 200, 220, 175, } ,
		    0, 0, L"", 8, L"Tahoma"},
		{

		/* Group box */
		{{WS_CHILD | WS_VISIBLE | BS_GROUPBOX, 0, 5, 5, 210, 150,
			 ID_GROUP}, 0xffff, 0x80, L"Settings   ", 0},
		
		/* Listening port */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 15, 40, 12, ID_STATIC},
			0xffff, 0x082, L"Port       ", 0},
		{{WS_CHILD | WS_VISIBLE | WS_BORDER | WS_TABSTOP |
			ES_AUTOHSCROLL, 0L, 10, 15, 145, 12, ID_EDIT_PORT},
			0xffff, 0x081, L"           ", 0},

		/* Document root */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 30, 40, 12, ID_STATIC},
			0xffff, 0x082, L"WWW root   ", 0},
		{{WS_CHILD | WS_VISIBLE | WS_BORDER | WS_TABSTOP |
			ES_AUTOHSCROLL, 0L, 10, 30, 145, 12, ID_EDIT_WWWROOT},
			0xffff, 0x081, L"           ", 0},

		/* Index file */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 45, 40, 12, ID_STATIC},
			0xffff, 0x082, L"Index file ", 0},
		{{WS_CHILD | WS_VISIBLE | WS_BORDER | WS_TABSTOP |
			ES_AUTOHSCROLL, 0L, 10, 45, 145, 12, ID_EDIT_INDEX},
			0xffff, 0x081, L"           ", 0},

		/* CGI pattern */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 60, 40, 12, ID_STATIC},
			0xffff, 0x082, L"CGI pattern", 0},
		{{WS_CHILD | WS_VISIBLE | WS_BORDER | WS_TABSTOP |
			 ES_AUTOHSCROLL, 0L, 10, 60, 145, 12, ID_EDIT_CGIPAT},
			0xffff, 0x081, L"           ", 0},

		/* Access Log */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 75, 40, 12, ID_STATIC},
			0xffff, 0x082, L"Access log ", 0},
		{{WS_CHILD | WS_VISIBLE | WS_BORDER | WS_TABSTOP |
			 ES_AUTOHSCROLL, 0L, 10, 75, 145, 12, ID_EDIT_LOG},
			0xffff, 0x081, L"           ", 0},

		/* Error Log */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 90, 40, 12, ID_STATIC},
			0xffff, 0x082, L"Error log  ", 0},
		{{WS_CHILD | WS_VISIBLE | WS_BORDER | WS_TABSTOP |
			 ES_AUTOHSCROLL, 0L, 10, 90, 145, 12, ID_EDIT_ELOG},
			0xffff, 0x081, L"           ", 0},

		/* Do directory listing */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 105, 40, 12, ID_STATIC},
			0xffff, 0x082, L"List dirs  ", 0},
		{{WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX | WS_TABSTOP,
			0L, 10, 105, 145, 12, ID_CB_DIRLIST},
			0xffff, 0x080, L"           ", 0},
		
		/* Certificate File */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 120, 40, 12, ID_STATIC},
			0xffff, 0x082, L"SSL certif ", 0},
		{{WS_CHILD | WS_VISIBLE | WS_BORDER | WS_TABSTOP |
			 ES_AUTOHSCROLL, 0L, 10, 120, 145, 12, ID_EDIT_CERT},
			0xffff, 0x081, L"           ", 0},

		/* Do SSL */
		{{WS_CHILD | WS_VISIBLE, 0L, 165, 135, 40, 12, ID_STATIC},
			0xffff, 0x082, L"Use SSL    ", 0},
		{{WS_CHILD | WS_VISIBLE | BS_AUTOCHECKBOX | WS_TABSTOP,
			0L, 10, 135, 145, 12, ID_CB_DOSSL},
			0xffff, 0x080, L"           ", 0},
		
		/* Run button. */
		{{WS_CHILD | WS_VISIBLE | BS_PUSHBUTTON | WS_TABSTOP, 0L,
			165, 160, 50, 12, ID_BTN_RUN},
			0xffff,	0x0080, L"Run        ", 0},
		}
	};

	logfile = "c:\\shttpd.log";
	errfile= "c:\\error.log";
	docroot = "c:\\";

	DialogBoxIndirect(inst, (LPDLGTEMPLATE) &dialog, NULL, DlgProc);
}
#endif /* WITH_GUI */
#endif /* _WIN32 */

/*
 * Setup listening socket on given port, return socket
 */
static int
Listen(uint16_t port)
{
	int		sock, on = 1, af;
	struct sa	sa;

#ifdef WITH_IPV6
	af = PF_INET6;
#else
	af = PF_INET;
#endif /* WITH_IPV6 */

	if ((sock = socket(af, SOCK_STREAM, 6)) == -1)
		elog(1, "Listen: socket: %s", strerror(ERRNO));

	/* Make it nonblock to avoid block in accept() */
	(void) nonblock(sock);
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,(char *) &on, sizeof(on));
	
#ifdef WITH_IPV6
	sa.u.sin6.sin6_family = af;
	sa.u.sin6.sin6_port = htons(port);
	sa.u.sin6.sin6_addr = in6addr_any;
	sa.len = sizeof(sa.u.sin6);
#else
	sa.u.sin.sin_family = af;
	sa.u.sin.sin_port = htons(port);
	sa.u.sin.sin_addr.s_addr = INADDR_ANY;
	sa.len = sizeof(sa.u.sin);
#endif /* WITH_IPV6 */

	if (bind(sock, &sa.u.sa, sa.len) < 0)
		elog(1, "Listen: af %d bind(%d): %s", af, port,strerror(ERRNO));

	(void) listen(sock, 128);

	return (sock);
}

/*
 * Wrapper around open(), that takes care about directory separators
 */
static int
myopen(const char *path, int mode)
{
	char	buf[PATHSIZ];
	size_t	i;

	/* Copy the string, substituting the directory separator */
	for (i = 0; i < sizeof(buf) - 1 && path[i] != '\0'; i++)
		buf[i] = path[i] == '/' ? DIRSEP : path[i];
	buf[i] = '\0';

	return (open(buf, mode));
}

/*
 * The wrapper around stat(), that takes care about directory separators
 */
static int
mystat(const char *path, struct stat *stp)
{
	char	buf[PATHSIZ];
	size_t	i;

	/* Copy the string, substituting the directory separator */
	for (i = 0; i < sizeof(buf) - 1 && path[i] != '\0'; i++)
		buf[i] = path[i] == '/' ? DIRSEP : path[i];

	/* Trim last directory separator if any*/
	if (i > 0 && buf[i - 1] == DIRSEP)
		i--;
	buf[i] = '\0';

	return (stat(buf, stp));
}

/*
 * 'Reader' function for directory listing
 */
static int
directory_read(void *private, char *buf, size_t buflen)
{
	struct dir_data	*data = private;
	struct dirent	*dp = NULL;
	char		file[PATHSIZ], line[PATHSIZ], size[64], modified[64];
	const char	*slash;
	struct stat	st;
	size_t		n, total = 0;

	elog(0, "directory_read called %p %p %u", private, buf, buflen);

	if (buflen < sizeof(file)) {
		errno = EWOULDBLOCK;
		return (-1);
	}

	slash = data->path[strlen(data->path) - 1] == '/' ? "" : "/";
	
	while (buflen > sizeof(file) && (dp = readdir(data->dirp)) != NULL) {

		/* Do not show current dir and passwords file */
		if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, htpass))
			continue;

		(void) snprintf(file, sizeof(file), "%s%s%s",
		    data->path, slash, dp->d_name);
		
		(void) mystat(file, &st);
		
		if (st.st_mode & S_IFDIR) {
			(void) snprintf(size, sizeof(size), "%s", "[DIR]");
		} else {
			(void) snprintf(size, sizeof(size), "%llu", st.st_size);
		}
		(void) strftime(modified, sizeof(modified),
		    "%d-%b-%Y %H:%M", localtime(&st.st_mtime));

		n = snprintf(line, sizeof(line),
		    "<tr><td><a href=\"%s\">%s%s</a></td>"
		    "<td>&nbsp;&nbsp;%s</td>"
		    "<td>&nbsp;&nbsp;&nbsp;&nbsp;%s</td></tr>\n",
		    file + 1, dp->d_name,
		    st.st_mode & S_IFDIR ? "/" : "", modified, size);
		
		(void) memcpy(buf + total, line, n);
		total += n;
		buflen -=n;
	}

	if (dp)
		errno = EWOULDBLOCK;
	
	return (total);
}

/*
 * Directory listing cleanup function
 */
static void
directory_cleanup(void *private)
{
	struct dir_data	*data = private;

	if (data != NULL) {
		(void) closedir(data->dirp);
		free(data->path);
		free(data);
	}
}

/*
 * Directory was requested, setup to return listing
 */
static void
do_directory(struct conn *c, const char *file)
{
	struct dir_data	*data;
	struct iobuf	*io = &c->local;

	if ((data = calloc(1, sizeof(*data))) == NULL) {
		elog(0, "directory: calloc: %s", strerror(errno));
		return;
	} else if ((data->dirp = opendir(file)) == NULL) {
		elog(0, "directory: opendir(%s): %s", file, strerror(errno));
		free(data);
		return;
	}

	/* Read buffer should be embty */
	assert(io->nread == 0);

	io->nread = snprintf(io->buf + io->nread, sizeof(io->buf) - io->nread,
	    "HTTP/1.1 200 OK\r\n"
	    "Content-Type: text/html\r\n"
	    "\r\n"
	    "<html><head><title>Index of %s</title></head>"
	    "<body><h1>Index of %s</h1><table>"
	    "<tr><th>Name</th><th>Modified</th><th>Size</th></tr>",
	    file + 1, file + 1);

	io->read.iofunc		= directory_read;
	io->read.cleanup	= directory_cleanup;
	io->read.private	= data;
	data->path		= strdup(file);
	
	accesslog(c, 200);
}

/*
 * Read from socket
 */
static int
socket_read(void *private, char *buf, size_t buflen)
{
	elog(0, "socket_read: called %p %p %u", private, buf, buflen);
	return (recv((int) private, buf, buflen, 0));
}

/*
 * Write to socket
 */
static int
socket_write(void *private, char *buf, size_t buflen)
{
	elog(0, "socket_write: called %p %p %u", private, buf, buflen);
	return (send((int) private, buf, buflen, 0));
}

/*
 * Socket cleanup
 */
static void
socket_cleanup(void *private)
{
	elog(0, "socket_cleanup: called %p", private);

	if (private != NULL)
		(void) closesocket((int) private);
}

/*
 * Read from file
 */
static int
file_read(void *private, char *buf, size_t buflen)
{
	elog(0, "file_read: called %p %p %u", private, buf, buflen);
	return (read((int) private, buf, buflen));
}

#ifndef NO_CGI
/*
 * Write to file
 */
static int
file_write(void *private, char *buf, size_t buflen)
{
	elog(0, "file_write: called %p %p %u", private, buf, buflen);
	return (write((int) private, buf, buflen));
}
#endif /* NO_CGI */

/*
 * File cleanup
 */
static void
file_cleanup(void *private)
{
	elog(0, "file_cleanup: called %p", private);
	
	if (private != NULL)
		(void) close((int) private);
}


/*
 * Read from SSL-ed socket
 */
static int
ssl_read(void *private, char *buf, size_t buflen)
{
	elog(0, "ssl_read: called %p %p %u", private, buf, buflen);
	return (SSL_read(private, buf, buflen));
}

/*
 * Write to file
 */
static int
ssl_write(void *private, char *buf, size_t buflen)
{
	elog(0, "ssl_write: called %p %p %u", private, buf, buflen);
	return (SSL_write(private, buf, buflen));
}

/*
 * SSL cleanup
 */
static void
ssl_cleanup(void *private)
{
	elog(0, "ssl_cleanup: called %p", private);
	if (private) {
		(void) SSL_shutdown(private);
		(void) closesocket(SSL_get_fd(private));
		SSL_free(private);
	}
}

static void
closeiobuf(struct iobuf *io)
{
	if (io->read.cleanup)
		io->read.cleanup(io->read.private);

	if (io->shared == 0 && io->write.cleanup)
		io->write.cleanup(io->write.private);

	io->read.iofunc = io->write.iofunc = NULL;
	io->read.private = io->write.private = NULL;
}

/*
 * Close connection
 */
static void
closeconn(struct conn *c)
{
	closeiobuf(&c->remote);
	closeiobuf(&c->local);
}

/*
 * Convert month to the month number. Return 1 if OK, 0 if error
 */
static int
montoi(const char *s, int *number)
{
	int	num = -1;
	
	if (!casecmp(s, "Jan") || !casecmp(s, "January"))
		num = 0;
	else if (!casecmp(s, "Feb") || !casecmp(s, "February"))
		num = 1;
	else if (!casecmp(s, "Mar") || !casecmp(s, "March"))
		num = 2;
	else if (!casecmp(s, "Apr") || !casecmp(s, "April"))
		num = 3;
	else if (!casecmp(s, "May"))
		num = 4;
	else if (!casecmp(s, "Jun") || !casecmp(s, "June"))
		num = 5;
	else if (!casecmp(s, "Jul") || !casecmp(s, "July"))
		num = 6;
	else if (!casecmp(s, "Aug") || !casecmp(s, "August"))
		num = 7;
	else if (!casecmp(s, "Sep") || !casecmp(s, "September"))
		num = 8;
	else if (!casecmp(s, "Oct") || !casecmp(s, "October"))
		num = 9;
	else if (!casecmp(s, "Nov") || !casecmp(s, "November"))
		num = 10;
	else if (!casecmp(s, "Dec") || !casecmp(s, "December"))
		num = 11;

	*number = num;
	
	return (num != -1);
}

/*
 * Parse date-time string, and return the corresponding time_t value
 */
static time_t
datetosec(const struct vec *v)
{
	struct tm	tm;
	char		mon[32];
	int		sec, min, hour, mday, month, year;

	(void) memset(&tm, 0, sizeof(tm));

	/* DD/month/YY HH:MM:SS */
	if (((sscanf((char *) v->ptr, "%d/%32[a-zA-Z]/%d %d:%d:%d",
	    &mday, mon, &year, &hour, &min, &sec) == 6) ||
	    (sscanf((char *) v->ptr, "%d %32[a-zA-Z] %d %d:%d:%d",
	    &mday, mon, &year, &hour, &min, &sec) == 6) ||
	    (sscanf((char *) v->ptr, "%d-%32[a-zA-Z]-%d %d:%d:%d",
	    &mday, mon, &year, &hour, &min, &sec) == 6))
	    && montoi(mon, &month)) {
		tm.tm_mday	= mday;
		tm.tm_mon	= month;
		tm.tm_year	= year;
		tm.tm_hour	= hour;
		tm.tm_min	= min;
		tm.tm_sec	= sec;
	}
	
	if (tm.tm_year > 1900)
		tm.tm_year -= 1900;
	else if (tm.tm_year < 70)
		tm.tm_year += 100;

	elog(0, "datetosec: %u", mktime(&tm));

	return (mktime(&tm));
}

/*
 * Callback function called fow every header when request is parsed.
 * the connection is passwd as arg
 */
static void
cbheader(const struct hthdr *header, void *arg)
{
	struct vec	auth = {13, "Authorization"},
			cookie = {6, "Cookie"},
			cl = {14, "Content-Length"},
			loc = {8, "Location"},
			status = {6, "Status"},
			ims = {17, "If-Modified-Since"},
			ua = {10, "User-Agent"},
			ref = {7, "Referer"},
			ct = {12, "Content-Type"};
	struct conn	*c = arg;
	
#if 0
	elog(0, "cbheader: %.*s %.*s",
	    header->name.len, (char*)header->name.ptr,
	    header->value.len, (char*)header->value.ptr);
#endif

	if (vcasecmp(&header->name, &auth))
		c->auth = header->value;
	else if (vcasecmp(&header->name, &cookie))
		c->cookie = header->value;
	else if (vcasecmp(&header->name, &cl))
		c->clength = header->value;
	else if (vcasecmp(&header->name, &ct))
		c->ctype = header->value;
	else if (vcasecmp(&header->name, &loc))
		c->location = header->value;
	else if (vcasecmp(&header->name, &status))
		c->status = header->value;
	else if (vcasecmp(&header->name, &ua))
		c->useragent = header->value;
	else if (vcasecmp(&header->name, &ref))
		c->referer = header->value;
	else if (vcasecmp(&header->name, &ims))
		c->ims = datetosec(&header->value);
}

/*
 * Send an error back to a client.
 */
static void
senderr(struct conn *c, int status, const char *descr,
		const char *headers, const char *fmt, ...)
{
	char	msg[IOBUFSIZ];
	va_list	ap;
	int	n;

	n = snprintf(msg, sizeof(msg), "HTTP/1.1 %d %s\r\n%s%s\r\n%d ",
	    status, descr, headers, headers[0] == '\0' ? "" : "\r\n", status);

	va_start(ap, fmt);
	n += vsnprintf(msg + n, sizeof(msg) - n, fmt, ap);
	va_end(ap);

	c->remote.write.iofunc(c->remote.write.private, msg, n);

	accesslog(c, status);

	c->remote.done++;
}

#ifndef NO_CGI
/*
 * Open program, and return read/write pipes to it
 */
static int
redirect(const char *prog, int *in, int *out, char *envblk, char **envp)
{
	int	r[2], w[2];

#ifdef _WIN32
	HANDLE			pid;
	char			cmdline[2048], line[1024] = "#!", *p;
	FILE			*fp;
	STARTUPINFO		si;
	PROCESS_INFORMATION	pi;

	if (_pipe(r, BUFSIZ, _O_BINARY) == -1 || _pipe(w, BUFSIZ, _O_BINARY)) {
		elog(0, "redirect: _pipe");
		return (-1);
	}

	(void) memset(&si, 0, sizeof(si));
	si.cb = sizeof(si);
	si.dwFlags = STARTF_USESHOWWINDOW | STARTF_USESTDHANDLES;
	si.wShowWindow = SW_HIDE;
	si.hStdInput = (HANDLE) _get_osfhandle(w[0]);
	si.hStdOutput = si.hStdError = (HANDLE) _get_osfhandle(r[1]);

	/* If CGI file is a script, try to read the interpreter line */
	if ((fp = fopen(prog, "r")) != NULL) {
		(void) fgets(line, sizeof(line), fp);
		if (line[0] != '#' || line[1] != '!')
			line[2] = '\0';
		/* Trim whitespaces from interpreter name */
		for (p = &line[strlen(line) - 1]; p > line && isspace(*p); p--)
			*p = '\0';
		
		(void) fclose(fp);
	}

	/* Prepare command line */
	(void) snprintf(cmdline, sizeof(cmdline), "%s%s%s",
	    line + 2, line[2] == '\0' ? "" : " ", prog);
	elog(0, "redirect: cmdline [%s]", cmdline);

	if (CreateProcess(NULL, cmdline, NULL, NULL, TRUE,
	    CREATE_NEW_PROCESS_GROUP, envblk, NULL, &si, &pi) == 0) {
		elog(0, "redirect: CreateProcess(%s): %d", cmdline, ERRNO);
		_close(r[0]), _close(r[1]), _close(w[0]), close(w[1]);
		return (-1);
	}

	pid = pi.hProcess;
	
	(void) _close(r[1]);
	(void) _close(w[0]);
	
	*in = r[0];
	*out = w[1];
	
	return (0);
#else
	pid_t		pid;
	unsigned	i, word;
	char		*argv[128], *s;

	envblk = NULL;	/* unused */

	/* Prepare the argv */
	for (i = word = 0, s = strdup(prog); s && *s && i < NELEMS(argv); s++)
		if (isspace(*s) && word) {
			*s = '\0';
			word = 0;
		} else if (!isspace(*s) && !word) {
			argv[i++] = s;
			word = 1;
		}
	argv[i] = NULL;

	if (pipe(r) != 0 || pipe(w) != 0) {
		elog(0, "redirect: pipe");
		return (-1);
	}

	if ((pid = fork()) == -1) {
		elog(0, "redirect: fork");
		return (-1);
	} else if (pid == 0) {
		/* Child */
		close(r[0]), close(w[1]), close(0), close(1);
		if (dup2(r[1], 1) == -1 || dup2(w[0], 0) == -1)
			elog(1, "redirect: dup2");

		(void) execve(argv[0], argv, envp);
		elog(1, "redirect: exec(%s)", argv[0]);	/* Error, exit*/
	} else {
		/* Parent */
		(void) close(r[1]);
		(void) close(w[0]);
		*in = r[0];
		*out = w[1];
	}

	return (0);
#endif /* _WIN32 */
}

/*
 * Prepare the environment for the CGI program, and start CGI program.
 */
static int
spawncgi(struct conn *c, const char *prog, const char *pathinfo)
{
	int	n = 0, k = 0, ret, fdread, fdwrite;
	char	path[PATHSIZ], env[ENVSIZ], *penv[32], *p;

#define	ADDENV(x)		\
	penv[k++] = &env[n], n += 1 + snprintf(env + n, sizeof(env) - n, x)
#define	ADDENV1(x, y)		\
	penv[k++] = &env[n], n += 1 + snprintf(env + n, sizeof(env) - n, x,y)
#define	ADDENV2(x, y, z)	\
	penv[k++] = &env[n], n += 1 + snprintf(env + n, sizeof(env) - n, x,y,z)
	
	/* Prepare the environment block */
	ADDENV("GATEWAY_INTERFACE=CGI/1.1");
	ADDENV("SERVER_PROTOCOL=HTTP/1.1");
	ADDENV("REDIRECT_STATUS=1");			/* For PHP */
	ADDENV1("SERVER_PORT=%u", lport);
	ADDENV1("REMOTE_ADDR=%s", inet_ntoa(c->sa.u.sin.sin_addr));
	ADDENV2("HTTP_USER_AGENT=%.*s",
	    c->useragent.len, (char *) c->useragent.ptr);
	ADDENV2("SCRIPT_NAME=%.*s", c->uri.len, (char *) c->uri.ptr);
	ADDENV2("REQUEST_METHOD=%.*s",
	    c->hti.method.len, (char *) c->hti.method.ptr);

	/* Set up PAHT_INFO and PATH_TRANSLATED */
	if (pathinfo != NULL) {
		(void) snprintf(path, sizeof(path), "/%s", pathinfo);
		ADDENV1("PATH_INFO=%s", path);
	}
	(void) snprintf(path, sizeof(path), "%s/%s", cwd, prog);
	ADDENV1("PATH_TRANSLATED=%s", path);
	
	if ((p = getenv("PATH")) != NULL)
		ADDENV1("PATH=%s", p);
	
	if ((p = getenv("LD_LIBRARY_PATH")) != NULL)
		ADDENV1("LD_LIBRARY_PATH=%s",p);
	
	if (c->req.len > 0)
		ADDENV2("QUERY_STRING=%.*s", c->req.len, (char *) c->req.ptr);

	if (c->referer.len > 0)
		ADDENV2("HTTP_REFERER=%.*s",
		    c->referer.len, (char *) c->referer.ptr);

	if ((p = getenv("PERLLIB")) != NULL)
		ADDENV1("PERLLIB=%s", p);
	
	if (c->user[0] != '\0') {
		ADDENV1("REMOTE_USER=%s", c->user);
		ADDENV("AUTH_TYPE=Basic");
	}

	if (c->cookie.len > 0)
		ADDENV2("HTTP_COOKIE=%.*s",c->cookie.len,(char *)c->cookie.ptr);

	if (c->clength.len > 0)
		ADDENV1("CONTENT_LENGTH=%d", vtoint(&c->clength));

	if (c->ctype.len > 0)
		ADDENV2("CONTENT_TYPE=%.*s",c->ctype.len,(char *)c->ctype.ptr);

	penv[k++] = NULL;
	env[n++] = '\0';

	assert(k < (int) NELEMS(penv));
	assert(n < (int) NELEMS(env));

	ret = redirect(prog, &fdread, &fdwrite, env, penv);
	(void) nonblock(fdread);
	(void) nonblock(fdwrite);
	
	c->local.read.iofunc	= file_read;
	c->local.write.iofunc	= file_write;
	c->local.read.cleanup	= c->local.write.cleanup = file_cleanup;
	c->local.read.private	= (void *) fdread;
	c->local.write.private	= (void *) fdwrite;

	return (ret);
}
#endif /* NO_CGI */

/*
 * Parse CGI reply, construct reply line, and set gotreply flag so
 * further CGI output will be passed through
 */
static void
cgireply(struct conn *c)
{
	int	n, status = 200;
	char	buf[IOBUFSIZ], *p;

	if (htparse(c->local.buf, c->local.nread, &c->hti, cbheader, c) < 0)
		return;

	c->gotreply++;

	if (c->status.len > 0)
		status = vtoint(&c->status);
	else if (c->location.len > 0)
		status = 302;

	switch (status) {
	case 200:	p = "OK";		break;
	case 302:	p = "Found";		break;
	case 304:	p = "Not Modified";	break;
	case 400:	p = "Bad Request";	break;
	case 401:	p = "Unauthorized";	break;
	case 403:	p = "Forbidden";	break;
	case 404:	p = "Not Found";	break;
	case 408:	p = "Request Timeout";	break;
	case 500:	p = "Internal Error";	break;
	case 501:	p = "Not Implemented";	break;
	case 503:	p = "Service Temporarily Overloaded"; break;
	default:	p = "Shit Happened!";	break;
	}

	/* Reply line */
	n = snprintf(buf, sizeof(buf), "HTTP/1.1 %d %s\r\n", status, p);

	assert(c->remote.write.iofunc != NULL);
	
	c->remote.write.iofunc(c->remote.write.private, buf, n);
}

#ifndef NO_AUTH
/*
 * MD5 crypto algorithm.
 */
typedef struct _MD5Context {
	uint32_t	buf[4];
	uint32_t	bits[2];
	unsigned char	in[64];
} _MD5_CTX;

#if __BYTE_ORDER == 1234
#define byteReverse(buf, len)	/* Nothing */
#else
/*
 * Note: this code is harmless on little-endian machines.
 */
static void byteReverse(unsigned char *buf, unsigned longs)
{
	uint32_t t;
	do {
		t = (uint32_t) ((unsigned) buf[3] << 8 | buf[2]) << 16 |
			((unsigned) buf[1] << 8 | buf[0]);
		*(uint32_t *) buf = t;
		buf += 4;
	} while (--longs);
}
#endif /* __BYTE_ORDER */


/* The four core functions - F1 is optimized somewhat */

/* #define F1(x, y, z) (x & y | ~x & z) */
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f, w, x, y, z, data, s) \
( w += f(x, y, z) + data,  w = w<<s | w>>(32-s),  w += x )

/*
 * Start MD5 accumulation.  Set bit count to 0 and buffer to mysterious
 * initialization constants.
 */
static void _MD5Init(_MD5_CTX *ctx)
{
	ctx->buf[0] = 0x67452301;
	ctx->buf[1] = 0xefcdab89;
	ctx->buf[2] = 0x98badcfe;
	ctx->buf[3] = 0x10325476;

	ctx->bits[0] = 0;
	ctx->bits[1] = 0;
}


/*
 * The core of the MD5 algorithm, this alters an existing MD5 hash to
 * reflect the addition of 16 longwords of new data.  MD5Update blocks
 * the data and converts bytes into longwords for this routine.
 */
static void _MD5Transform(uint32_t buf[4], uint32_t const in[16])
{
	register uint32_t a, b, c, d;

	a = buf[0];
	b = buf[1];
	c = buf[2];
	d = buf[3];

	MD5STEP(F1, a, b, c, d, in[0] + 0xd76aa478, 7);
	MD5STEP(F1, d, a, b, c, in[1] + 0xe8c7b756, 12);
	MD5STEP(F1, c, d, a, b, in[2] + 0x242070db, 17);
	MD5STEP(F1, b, c, d, a, in[3] + 0xc1bdceee, 22);
	MD5STEP(F1, a, b, c, d, in[4] + 0xf57c0faf, 7);
	MD5STEP(F1, d, a, b, c, in[5] + 0x4787c62a, 12);
	MD5STEP(F1, c, d, a, b, in[6] + 0xa8304613, 17);
	MD5STEP(F1, b, c, d, a, in[7] + 0xfd469501, 22);
	MD5STEP(F1, a, b, c, d, in[8] + 0x698098d8, 7);
	MD5STEP(F1, d, a, b, c, in[9] + 0x8b44f7af, 12);
	MD5STEP(F1, c, d, a, b, in[10] + 0xffff5bb1, 17);
	MD5STEP(F1, b, c, d, a, in[11] + 0x895cd7be, 22);
	MD5STEP(F1, a, b, c, d, in[12] + 0x6b901122, 7);
	MD5STEP(F1, d, a, b, c, in[13] + 0xfd987193, 12);
	MD5STEP(F1, c, d, a, b, in[14] + 0xa679438e, 17);
	MD5STEP(F1, b, c, d, a, in[15] + 0x49b40821, 22);

	MD5STEP(F2, a, b, c, d, in[1] + 0xf61e2562, 5);
	MD5STEP(F2, d, a, b, c, in[6] + 0xc040b340, 9);
	MD5STEP(F2, c, d, a, b, in[11] + 0x265e5a51, 14);
	MD5STEP(F2, b, c, d, a, in[0] + 0xe9b6c7aa, 20);
	MD5STEP(F2, a, b, c, d, in[5] + 0xd62f105d, 5);
	MD5STEP(F2, d, a, b, c, in[10] + 0x02441453, 9);
	MD5STEP(F2, c, d, a, b, in[15] + 0xd8a1e681, 14);
	MD5STEP(F2, b, c, d, a, in[4] + 0xe7d3fbc8, 20);
	MD5STEP(F2, a, b, c, d, in[9] + 0x21e1cde6, 5);
	MD5STEP(F2, d, a, b, c, in[14] + 0xc33707d6, 9);
	MD5STEP(F2, c, d, a, b, in[3] + 0xf4d50d87, 14);
	MD5STEP(F2, b, c, d, a, in[8] + 0x455a14ed, 20);
	MD5STEP(F2, a, b, c, d, in[13] + 0xa9e3e905, 5);
	MD5STEP(F2, d, a, b, c, in[2] + 0xfcefa3f8, 9);
	MD5STEP(F2, c, d, a, b, in[7] + 0x676f02d9, 14);
	MD5STEP(F2, b, c, d, a, in[12] + 0x8d2a4c8a, 20);

	MD5STEP(F3, a, b, c, d, in[5] + 0xfffa3942, 4);
	MD5STEP(F3, d, a, b, c, in[8] + 0x8771f681, 11);
	MD5STEP(F3, c, d, a, b, in[11] + 0x6d9d6122, 16);
	MD5STEP(F3, b, c, d, a, in[14] + 0xfde5380c, 23);
	MD5STEP(F3, a, b, c, d, in[1] + 0xa4beea44, 4);
	MD5STEP(F3, d, a, b, c, in[4] + 0x4bdecfa9, 11);
	MD5STEP(F3, c, d, a, b, in[7] + 0xf6bb4b60, 16);
	MD5STEP(F3, b, c, d, a, in[10] + 0xbebfbc70, 23);
	MD5STEP(F3, a, b, c, d, in[13] + 0x289b7ec6, 4);
	MD5STEP(F3, d, a, b, c, in[0] + 0xeaa127fa, 11);
	MD5STEP(F3, c, d, a, b, in[3] + 0xd4ef3085, 16);
	MD5STEP(F3, b, c, d, a, in[6] + 0x04881d05, 23);
	MD5STEP(F3, a, b, c, d, in[9] + 0xd9d4d039, 4);
	MD5STEP(F3, d, a, b, c, in[12] + 0xe6db99e5, 11);
	MD5STEP(F3, c, d, a, b, in[15] + 0x1fa27cf8, 16);
	MD5STEP(F3, b, c, d, a, in[2] + 0xc4ac5665, 23);

	MD5STEP(F4, a, b, c, d, in[0] + 0xf4292244, 6);
	MD5STEP(F4, d, a, b, c, in[7] + 0x432aff97, 10);
	MD5STEP(F4, c, d, a, b, in[14] + 0xab9423a7, 15);
	MD5STEP(F4, b, c, d, a, in[5] + 0xfc93a039, 21);
	MD5STEP(F4, a, b, c, d, in[12] + 0x655b59c3, 6);
	MD5STEP(F4, d, a, b, c, in[3] + 0x8f0ccc92, 10);
	MD5STEP(F4, c, d, a, b, in[10] + 0xffeff47d, 15);
	MD5STEP(F4, b, c, d, a, in[1] + 0x85845dd1, 21);
	MD5STEP(F4, a, b, c, d, in[8] + 0x6fa87e4f, 6);
	MD5STEP(F4, d, a, b, c, in[15] + 0xfe2ce6e0, 10);
	MD5STEP(F4, c, d, a, b, in[6] + 0xa3014314, 15);
	MD5STEP(F4, b, c, d, a, in[13] + 0x4e0811a1, 21);
	MD5STEP(F4, a, b, c, d, in[4] + 0xf7537e82, 6);
	MD5STEP(F4, d, a, b, c, in[11] + 0xbd3af235, 10);
	MD5STEP(F4, c, d, a, b, in[2] + 0x2ad7d2bb, 15);
	MD5STEP(F4, b, c, d, a, in[9] + 0xeb86d391, 21);

	buf[0] += a;
	buf[1] += b;
	buf[2] += c;
	buf[3] += d;
}

/*
 * Update context to reflect the concatenation of another buffer full
 * of bytes.
 */
static void
_MD5Update(_MD5_CTX *ctx, unsigned char const *buf, unsigned len)
{
	uint32_t t;

	/* Update bitcount */

	t = ctx->bits[0];
	if ((ctx->bits[0] = t + ((uint32_t) len << 3)) < t)
		ctx->bits[1]++;		/* Carry from low to high */
	ctx->bits[1] += len >> 29;

	t = (t >> 3) & 0x3f;	/* Bytes already in shsInfo->data */

	/* Handle any leading odd-sized chunks */

	if (t) {
		unsigned char *p = (unsigned char *) ctx->in + t;

		t = 64 - t;
		if (len < t) {
			memcpy(p, buf, len);
			return;
		}
		memcpy(p, buf, t);
		byteReverse(ctx->in, 16);
		_MD5Transform(ctx->buf, (uint32_t *) ctx->in);
		buf += t;
		len -= t;
	}
	/* Process data in 64-byte chunks */

	while (len >= 64) {
		memcpy(ctx->in, buf, 64);
		byteReverse(ctx->in, 16);
		_MD5Transform(ctx->buf, (uint32_t *) ctx->in);
		buf += 64;
		len -= 64;
	}

	/* Handle any remaining bytes of data. */

	memcpy(ctx->in, buf, len);
}

/*
 * Final wrapup - pad to 64-byte boundary with the bit pattern 
 * 1 0* (64-bit count of bits processed, MSB-first)
 */
static void
_MD5Final(unsigned char digest[16], _MD5_CTX *ctx)
{
	unsigned count;
	unsigned char *p;

	/* Compute number of bytes mod 64 */
	count = (ctx->bits[0] >> 3) & 0x3F;

	/* Set the first char of padding to 0x80.  This is safe since there is
	   always at least one byte free */
	p = ctx->in + count;
	*p++ = 0x80;

	/* Bytes of padding needed to make 64 bytes */
	count = 64 - 1 - count;

	/* Pad out to 56 mod 64 */
	if (count < 8) {
		/* Two lots of padding:  Pad the first block to 64 bytes */
		memset(p, 0, count);
		byteReverse(ctx->in, 16);
		_MD5Transform(ctx->buf, (uint32_t *) ctx->in);

		/* Now fill the next block with 56 bytes */
		memset(ctx->in, 0, 56);
	} else {
		/* Pad block to 56 bytes */
		memset(p, 0, count - 8);
	}
	byteReverse(ctx->in, 14);

	/* Append length in bits and transform */
	((uint32_t *) ctx->in)[14] = ctx->bits[0];
	((uint32_t *) ctx->in)[15] = ctx->bits[1];

	_MD5Transform(ctx->buf, (uint32_t *) ctx->in);
	byteReverse((unsigned char *) ctx->buf, 4);
	memcpy(digest, ctx->buf, 16);
	memset((char *) ctx, 0, sizeof(ctx));	/* In case it's sensitive */
}

/*
 * Stringify binary data. Output buffer must be twice as big as input,
 * because each byte takes 2 bytes in string representation
 */
static void
bin2str(char *to, const unsigned char *p, size_t len)
{
	const char	*hex = "0123456789abcdef";
	
	for (;len--; p++) {
		*to++ = hex[p[0] >> 4];
		*to++ = hex[p[0] & 0x0f];
		
	}
}


/*
 * Return stringified MD5 hash for list of vectors.
 * buf must point to at least 32-bytes long buffer
 */
static void
md5(char *buf, ...)
{
	char		hash[16];
	struct vec	*v;
	va_list		ap;
	_MD5_CTX	ctx;

	_MD5Init(&ctx);

	va_start(ap, buf);
	while ((v = va_arg(ap, struct vec *)) != NULL)
		_MD5Update(&ctx, v->ptr, v->len);
	va_end(ap);

	_MD5Final(hash, &ctx);

	bin2str(buf, hash, sizeof(hash));
}

/*
 * HTTP digest authentication
 */
struct digest {
	struct vec	user;
	struct vec	nonce;
	struct vec	resp;
	struct vec	uri;
};

/*
 * Fetch a password provided by user.
 * Return 1 if success, 0 otherwise.
 */
static int
getauth(struct conn *c, struct digest *dig)
{
	const char	*p = (char *) c->auth.ptr, *e = p + c->auth.len;
	struct vec	*v = NULL;

	if (c->auth.len >= 7 && memcmp(p, "Digest ", 7) != 0)
		return (0);

	(void) memset(dig, 0, sizeof(dig));

#define	SETDIGVEC(N,PTR)	p += N - 1, v = PTR, v->ptr = p + 1, v->len = 0
	
	/* Parse the authentication header. TODO check memcmp overflow */
	for (p += 7; p < e; p++)
		if (*p == '"') {
			v = NULL;
		} else if (v) {
			v->len++;
		} else if (memcmp(p, "username=\"", 10) == 0) {
			SETDIGVEC(10, &dig->user);
		} else if (memcmp(p, "nonce=\"", 7) == 0) {
			SETDIGVEC(7, &dig->nonce);
		} else if (memcmp(p, "response=\"", 10) == 0) {
			SETDIGVEC(10, &dig->resp);
		} else if (memcmp(p, "uri=\"", 5) == 0) {
			SETDIGVEC(5, &dig->uri);
		}

	/* Copy the username into the connection descriptor */
	if (dig->user.len > sizeof(c->user) - 1)
		dig->user.len = sizeof(c->user) - 1;
	(void) memcpy(c->user, dig->user.ptr, dig->user.len);
	c->user[dig->user.len] = '\0';

	return (1);
}

/*
 * Check the user's password, return 1 if OK
 */
static int
checkpass(const struct conn *c, const char *a1, const struct digest *dig)
{
	char		a2[32], resp[32];
	struct vec	d = {1, ":"}, va2 = {32, a2}, va1 = {0, a1};

	elog(0, "checkpass: %lu %lu", now, vtoint(&dig->nonce));

	/* nonce must be not too old, and uris must match */
	if (dig->uri.len != c->uri.len ||
	    memcmp(dig->uri.ptr, c->uri.ptr, c->uri.len) ||
	    dig->resp.len != 32 ||
	    now - vtoint(&dig->nonce) > 300)
		return (0);

	va1.len = strlen(a1);

	md5(a2, &c->hti.method, &d, &dig->uri, NULL);
	md5(resp, &va1, &d, &dig->nonce, &d, &va2, NULL);

	return (memcmp(resp, dig->resp.ptr, 32) == 0);
}

/*
 * Check authorization. Return 1 if not needed or authorized, 0 otherwise
 */
static int
checkauth(struct conn *c, const char *file)
{
	int		i, authorized = 1;
	const char	*p, *e;
	char		authfile[2048], line[512], user[64], ha1[64], realm[64];
	struct digest	digest;
	FILE		*fp;

	if (globhtpass) {
		/* Use global passwords file */
		(void) snprintf(authfile, sizeof(authfile), "%s", globhtpass);
	} else {
		/* Try to find .htpasswd in requested directory */
		for (p = file, e = p + strlen(p) - 1; e > p; e--)
			if (*e == '/')
				break;
		assert(*e == '/');
		(void) snprintf(authfile, sizeof(authfile), "%.*s/%s",
		    e - p, p, htpass);
	
		/* Fix directory separators */
		for (i = 0; authfile[i] != '\0'; i++)
			if (authfile[i]== '/')
				authfile[i] = DIRSEP;
	}

	/* Try to open passwords file. If opened, set authorized to 0  */
	if ((fp = fopen(authfile, "r")) != NULL &&
	    !(authorized = 0) && getauth(c, &digest)) {
		
		/* Each line should be "USER:REALM:HASH" */
		while (fgets(line, sizeof(line), fp) != NULL)
			if (sscanf(line, "%64[^:]:%64[^:]:%64s",
			    user, realm, ha1) == 3 &&
			    !strcmp(c->user, user) &&
			    !strcmp(domain, realm)) {
				authorized = checkpass(c, ha1, &digest);
				break;
			}
		(void) fclose(fp);
	}

	return (authorized);
}
#endif /* NO_AUTH */

/*
 * Protect from ..
 */
static void
dotdot(char* file)
{
	char	*p, *s;
	int	n;

	/* Collapse any multiple / sequences. */
	while ((p = strstr(file, "//")) != NULL) {
		for (s = p + 2; *s == '/'; ++s)
			continue;
		(void) strcpy(p + 1, s);
	}

	/* Remove leading ./ and any /./ sequences. */
	while (strncmp(file, "./", 2) == 0)
		(void) strcpy(file, file + 2);
	
	while ((p = strstr(file, "/./")) != NULL)
		(void) strcpy(p, p + 2);

	/* Alternate between removing leading ../ and removing xxx/../ */
	for (;;) {
		while (strncmp(file, "../", 3) == 0)
			(void) strcpy(file, file + 3);
		if ((p = strstr(file, "/../")) == NULL)
			break;
		for (s = p - 1; s >= file && *s != '/'; --s)
			continue;
		(void) strcpy(s + 1, p + 4);
	}

	/* Also elide any xxx/.. at the end. */
	while ((n = strlen(file)) > 3 &&
	    strcmp((p = file + n - 3 ), "/..") == 0) {
		for (s = p - 1; s >= file && *s != '/'; --s)
			continue;
		if (s < file)
			break;
		*s = '\0';
	}
}

/*
 * For given directory path, substitute it to valid index file.
 * Return 0 if index file has been found, -1 if not found
 */
static int
useindex(char *path, size_t maxpath)
{
	struct stat	st;
	char		try[2048], index[64];
	const char	*s = ifile;
	
	do {
		if (sscanf(s, "%64[^,]", index)) {
			(void) snprintf(try, sizeof(try), "%s%s", path, index);
			if (mystat(try, &st) == 0) {
				strncat(path, index, maxpath - strlen(path) -1);
				return (0);
			}
		}
		
		/* Move to the next index file */
		if ((s = strchr(s, ',')) != NULL)
			s++;
	} while (s != NULL);

	return (-1);
}

/*
 * Decode urlencoded string. from and to may point to the same location
 */
static void
urldecode(char *from, char *to)
{
	int	a, b;
#define	HEXTOI(x)  (x >= '0' && x <= '9' ? x - '0' : x - 'a')

	for (; *from != '\0'; ++to, ++from) {
		if (from[0] == '%' && isxdigit(from[1]) && isxdigit(from[2])) {
			a = tolower(from[1]);
			b = tolower(from[2]);
			*to = (HEXTOI(a) << 4) + HEXTOI(b);
			from += 2;
		} else {
			*to = *from;
		}
	}

	*to = '\0';
}

/*
 * stat() requested file, return 0 if OK, -1 if error.
 * If the file is given arguments using PATH_INFO mechanism,
 * initialize pathinfo pointer.
 */
static int
getpathinfo(char *path, struct stat *stp, char **pathinfo)
{
	char	*p;

	*pathinfo = NULL;

	if (mystat(path, stp) == 0)
		return (0);
	
	/* Strip directory parts of the path one by one */
	for (p = path + strlen(path) - 1; p > path + 2; p--)
		if (*p == '/') {
			*p = '\0';
			if (mystat(path, stp) == 0) {
				*pathinfo = p + 1;
				return (0);
			} else {
				*p = '/';
			}
		}

	return (-1);
}

/*
 * Schedule to process regular file request
 */
static void
do_file(struct conn *c, const struct stat *stp, int fd)
{
	char		date[64], lm[64], etag[64];
	const char	*mime = "text/html";		/* Default mime type */
	const char	*s = (char *) c->uri.ptr + c->uri.len;
	struct mimetype	*p;
	
	/* Figure out the mime type */
	for (p = mimetypes; p != NULL; p = p->next)
		if (c->uri.len > p->extlen &&
		    *(s - p->extlen - 1) == '.' &&
		    !ncasecmp(p->ext, s - p->extlen, p->extlen)) {
			mime = p->mime;
			break;
		}
			
	/* Prepare Etag, Date, Last-Modified headers */
	strftime(date, sizeof(date), "%d/%b/%Y %H:%M:%S", localtime(&now));
	strftime(lm, sizeof(lm), "%d/%b/%Y %H:%M:%S",localtime(&stp->st_mtime));
	(void) snprintf(etag, sizeof(etag), "%lx.%lx",
	    (unsigned long) stp->st_mtime, (unsigned long) stp->st_size);

	/* Local read buffer should be empty */
	assert(c->local.nread == 0);

	c->local.nread = snprintf(c->local.buf, sizeof(c->local.buf),
	    "HTTP/1.1 200 OK\r\n"
	    "Connection: close\r\n"
	    "Date: %s\r\n"
	    "Last-Modified: %s\r\n"
	    "Content-Type: %s\r\n"
	    "Content-Length: %lu\r\n"
	    "\r\n", date, lm, mime, (unsigned long) stp->st_size);

	/* Set up read channel */
	c->local.read.private	= (void *) fd;
	c->local.read.iofunc	= file_read;
	c->local.read.cleanup	= file_cleanup;
	
	accesslog(c, 200);
}

/*
 * Try to open requested file
 */
static void
handle(struct conn *c)
{
	struct stat	st;
	struct url	url;
	const char	*p;
	char		path[IOBUFSIZ], *pathinfo;
	int		fd;

	if (htparse(c->remote.buf, c->remote.nread, &c->hti, cbheader, c) < 0)
		return;

	c->gotrequest++;
	c->remote.nwritten = c->hti.totlen;
	c->nexpected = c->hti.totlen + vtoint(&c->clength);

	if (hturl(c->hti.url.ptr, c->hti.url.len, &url) <= 0) {
		senderr(c, 400, "Bad Request", "", "Bad request");
	} else if (url.uri.len <= 0 || ((char *) url.uri.ptr)[0] != '/') {
		senderr(c, 400, "Bad Request","", "Bad Filename");
	} else {
		c->uri = url.uri;
	
		elog(0, "handle: [%.*s]", c->uri.len, (char *) c->uri.ptr);

		/* Strip QUERY_STRING from URI */
		if (c->uri.len > 0 && (p = vchr(&c->uri, '?')) != NULL) {
			c->req.len = (char *) c->uri.ptr + c->uri.len - p - 1;
			c->req.ptr = p + 1;
			c->uri.len = p - (char *) c->uri.ptr;
		}

		(void) snprintf(path, sizeof(path), "./%.*s",
		    c->uri.len, (char *) c->uri.ptr);

		urldecode(path + 2, path + 2);
		dotdot(path + 1);

		/* If directory has been requested, substitute index file */	
		if (path[strlen(path) - 1] == '/' &&
		    useindex(path, sizeof(path) - 1) == -1 && !do_dir) {
			senderr(c, 403, "Forbidden","",
			    "Directory Listing Denied");
			return;
		}
		
		if (strstr(path, htpass)) {
			senderr(c, 403, "Forbidden","", "Permission Denied");
		} else if (getpathinfo(path, &st, &pathinfo) != 0 ||
		    ((st.st_mode & S_IFDIR) && pathinfo != NULL)) {
			elog(0, "handle: stat(%s): %d", path, ERRNO);
			senderr(c, 404, "Not Found","", "Not Found");
#ifndef NO_AUTH
		} else if (checkauth(c, path) != 1) {
			char	buf[256];

			(void) snprintf(buf, sizeof(buf),
			    "WWW-Authenticate: Digest realm=\"%s\", "
			    "nonce=\"%lu\"", domain, (unsigned long) now);

			senderr(c, 401, "Unauthorized",
			    buf, "Authentication required");
#endif /* NO_AUTH */
		} else if ((st.st_mode & S_IFDIR) && do_dir) {
			do_directory(c, path);
		} else if ((st.st_mode & S_IFDIR) && !do_dir) {
			elog(0, "handle: %s: Denied", path);
			senderr(c, 403, "Forbidden","",
			    "Directory listing denied");
#ifndef NO_CGI
		} else if (strstr(path, cgipat) != NULL) {
			int n;
			c->cgimode++;
			if ((n = spawncgi(c, path, pathinfo)) != 0) {
				senderr(c, 500, "Server Error", "",
				    "Error executing CGI script");
				elog(0, "handle: could not exec: %d", ERRNO);
			}
			accesslog(c, n == 0 ? 200 : 500);
#endif /* NO_CGI */
		} else if (c->ims && c->ims >= st.st_mtime) {
			elog(0, "handle: %s: Not Modified", path);
			senderr(c, 304, "Not Modified","", "");
		} else if ((fd = myopen(path, O_RDONLY | O_BINARY)) == -1) {
			elog(0, "handle: open(%s): %s", path, strerror(errno));
			senderr(c, 500, "Error","", "Some kind of error");
		} else {
			do_file(c, &st, fd);
		}
	}
}

/*
 * Perform SSL handshake
 */
static void
handshake(struct conn *c)
{
	int	n;
	SSL	*ssl = c->remote.read.private;

	assert(ssl != NULL);
	assert(c->sslaccepted == 0);

	if ((n = SSL_accept(ssl)) == 0) {
		n = SSL_get_error(ssl, n);
		if (n != SSL_ERROR_WANT_READ && n != SSL_ERROR_WANT_WRITE)
			closeiobuf(&c->remote);
		elog(0, "handshake: SSL_accept: %d", n);
	} else {
		elog(0, "handshake: SSL accepted");
		c->sslaccepted = 1;
	}
}

/*
 * Push data to client
 */
static void
push(struct iobuf *from, struct iobuf *to)
{
	int	n, buflen;
	char	*buf;

	buf	= from->buf + from->nwritten;
	buflen	= from->nread - from->nwritten;
	
	assert(buflen > 0);
	assert(to->write.iofunc != NULL);

	n = to->write.iofunc(to->write.private, buf, buflen);
#if 0
	elog(0, "push: %p %p %d %d", from->read.private,
	    to->write.private, buflen, n);
#endif
	if (n > 0) {
#if 0
		elog(0, "push: [%.*s]", n, from->buf + from->nwritten);
#endif
		active++;
		from->nwritten += n;
		assert(from->nwritten <= from->nread);
		if (from->nwritten == from->nread)
			from->nwritten = from->nread = 0;
	} else if (n <= 0 && ERRNO != EWOULDBLOCK) {
		to->done++;
	}
}

/*
 * Read incoming data
 */
static int
pull(struct iobuf *io)
{
	int	n, buflen;

	buflen = sizeof(io->buf) - io->nread;	/* Bytes left in buffer */
	assert(io->read.iofunc != NULL);
	assert(buflen > 0);

	n = io->read.iofunc(io->read.private, io->buf + io->nread, buflen);
#if 1
	elog(0, "pull: %p: %d read (%d)", io->read.private, n, ERRNO);
#endif

	if (n == 0 || (n < 0 && ERRNO != EWOULDBLOCK)) {
		io->done++;
	} else if (n > 0) {
		active++;
		io->nread += n;
	}
	
	return (n);
}

/*
 * Accept new connection
 */
static void
Accept(int lsn) {
	struct conn	*c;
	struct sa	sa;
	SSL		*ssl = NULL;
	int		sock;
			
	sa.len = sizeof(sa.u.sin);
			    
	if ((sock = accept(lsn, &sa.u.sa, &sa.len)) == -1) {
#if 0
		elog(0, "Accept: accept: %s", strerror(ERRNO));
#endif
	} else if (nonblock(sock) != 0) {
		elog(0, "Accept: nonblock: %s", strerror(ERRNO));
		(void) close(sock);
	} else if (dossl && (ssl = SSL_new(ctx)) == NULL) {
		elog(0, "Accept: SSL_new: %s", strerror(ERRNO));
		(void) close(sock);
	} else if (dossl && SSL_set_fd(ssl, sock) == 0) {
		elog(0, "Accept: SSL_set_fd: %s", strerror(ERRNO));
		(void) close(sock);
		SSL_free(ssl);
	} else if ((c = calloc(1, sizeof(*c))) == NULL) {
		(void) close(sock);
		elog(0, "Accept: calloc: %s", strerror(ERRNO));
	} else {
		LL_TAIL(&conns, &c->link);
		c->sa = sa;
		if (ssl) {
			c->remote.read.iofunc	= ssl_read;
			c->remote.write.iofunc	= ssl_write;
			c->remote.read.private = c->remote.write.private = ssl;
			c->remote.read.cleanup = c->remote.write.cleanup =
			    ssl_cleanup;
		} else {
			c->remote.read.iofunc	= socket_read;
			c->remote.write.iofunc	= socket_write;
			c->remote.read.private = c->remote.write.private =
			    (void *) sock;
			c->remote.read.cleanup = c->remote.write.cleanup =
			    socket_cleanup;
		}
		
		c->remote.shared = 1;
		c->expire = now + ttl;
		elog(0, "%p connected.", c);

		if (ssl)
			handshake(c);
	}
}

static void
disconnect(struct conn *c)
{
	if (c->cgimode && !c->gotreply && c->local.nread > 0)
		senderr(c, 500, "Server Error", "",
			"Premature end of script headers");

	elog(0, "%p disconnected.", c);
	closeconn(c);
	LL_DEL(&c->link);
	free(c);
}

/*
 * One cycle of main server loop
 */
static void
loop(int listen_socket)
{
	struct llhead	*lp, *tmp;
	struct conn	*c;

	now = time(0);

	/* Accept an incoming connection if any */
	Accept(listen_socket);

	active = 0;

	/* Loop through all connections, handle if needed */
	LL_FOREACH_SAFE(&conns, lp, tmp) {
		c = LL_ENTRY(lp, struct conn, link);

#define	READSPACE(io)		((io)->nread < sizeof((io)->buf))
#define	DATAPRESENT(io)		((io)->nread > (io)->nwritten)

		if (dossl && c->sslaccepted == 0) {
			handshake(c);
		} else if (READSPACE(&c->remote) && pull(&c->remote) > 0) {
			if (!c->gotrequest)
				handle(c);
			c->expire = now + ttl;
		}

		if (c->local.read.iofunc && c->local.done == 0 &&
		    READSPACE(&c->local) && pull(&c->local) > 0)
			c->expire = now + ttl;

		if (c->cgimode && c->gotreply == 0)
			cgireply(c);

		if (DATAPRESENT(&c->local) && (c->cgimode == 0 || c->gotreply))
			push(&c->local, &c->remote);

		if (DATAPRESENT(&c->remote) && c->local.write.iofunc)
			push(&c->remote, &c->local);
		
		/* Disconnect 'done' and timed out clients */
		if (c->remote.done ||
		    (c->local.done && c->local.nread == 0) ||
		    c->expire < now)
			disconnect(c);
	}

	/* Sleep now if no active connections */
	if (active == 0)
		(void) usleep(100 * 1000);
}

/*
 * Redirect error log to a file
 */
static void
errtofile(const char *filename)
{
	int fd;
	
	if ((fd = open(filename, O_WRONLY | O_APPEND | O_CREAT, 0666)) == -1)
		elog(1, "open(%s): %s", errfile, strerror(ERRNO));

	(void) dup2(fd, 2);
	(void) close(fd);
}

/*
 * Add mime type to the known mime type list
 */
static void
addmimetype(const char *ext, const char *mime)
{
	struct mimetype	*p;

	/* XXX No error checking */
	p = calloc(1, sizeof(*p));
	assert(p != NULL);
	
	p->ext = strdup(ext);
	p->mime = strdup(mime);

	assert(p->ext != NULL);
	assert(p->mime != NULL);

	p->extlen = strlen(p->ext);
		
	p->next = mimetypes;
	mimetypes = p;
}

/*
 * Set mime types from file
 */
static void
setmimetypes(const char *file)
{
	FILE	*fp;
	char	line[512], ext[sizeof(line)], mime[sizeof(line)];
	
	if ((fp = fopen(file, "r")) == NULL)
		elog(1, "setmimetypes: fopen(%s): %s", file, strerror(errno));

	while (fgets(line, sizeof(line), fp) != NULL)
		if (sscanf(line, "%s %s", ext, mime) == 2)
			addmimetype(ext, mime);

	(void) fclose(fp);
}

/*
 * Set default mime types
 */
static void
setdefmimetypes(void)
{
	addmimetype("jpeg", "image/jpeg");
	addmimetype("png", "image/png");
	addmimetype("mpg", "video/mpeg");
	addmimetype("asf", "video/x-ms-asf");
	addmimetype("avi", "video/x-msvideo");
	addmimetype("bmp", "image/bmp");
	addmimetype("doc", "application/msword");
	addmimetype("exe", "application/octet-stream");
	addmimetype("zip", "application/zip");
	addmimetype("jpg", "image/jpeg");
	addmimetype("gif", "image/gif");
	addmimetype("txt", "text/plain");
	addmimetype("ico", "image/x-icon");
	addmimetype("css", "text/css");
	addmimetype("htm", "text/html");
	addmimetype("html", "text/html");
}

static void
doit(void)
{
	int	lsn;

	/* Init global variable 'now', needed for elog() etc */
	now = time(0);

	/* Redirect stderr to a file if requested */
	if (errfile)
		errtofile(errfile);
	
	/* Set mime types */
	if (mimefile)
		setmimetypes(mimefile);
	else
		setdefmimetypes();

	/* Open access log file */
	logfp = fopen(logfile, "a");

	/* Go to document root */
	if (chdir(docroot) != 0)
		elog(1, "chdir(%s): %s", docroot, strerror(ERRNO));
	if (getcwd(cwd, sizeof(cwd)) == NULL)
		elog(1, "getcwd: %s", strerror(ERRNO));

#ifdef _WIN32
	{WSADATA data; WSAStartup(MAKEWORD(2,2), &data);}
#endif /* _WIN32 */

	lsn = Listen(lport);
	
	/* Switch to alternate UID, it is safe now, after Listen() */
	if (uid > 0 && setuid(uid) == -1)
		elog(1, "main: setuid(%d): %s", uid, strerror(ERRNO));

	(void) signal(SIGTERM, sigterm);
	(void) signal(SIGINT, sigterm);
	(void) signal(SIGCHLD, sigchild);
	(void) signal(SIGPIPE, SIG_IGN);

	if (dossl) {
		/* Initialize SSL crap */
		SSL_load_error_strings();
		SSLeay_add_ssl_algorithms();
		if ((ctx = SSL_CTX_new(SSLv23_server_method())) == NULL)
			elog(1, "main: error: SSL_CTX_new %p", ctx);
		else if (SSL_CTX_use_certificate_file(ctx, certfile,
		    SSL_FILETYPE_PEM) == 0)
			elog(1, "main: please put %s here", certfile);
		else if (SSL_CTX_use_PrivateKey_file(ctx, certfile,
		    SSL_FILETYPE_PEM) == 0)
			elog(1, "main: please put %s here", certfile);
	}
	
	(void) printf("shttpd version %s (%s) started on port %d\n",
	    VERSION,
#ifdef WITH_SSL
	    "with SSL"
#else
	    "without SSL"
#endif /* WITH_SSL */
	    , lport);

	while (quit == 0)
		loop(lsn);

	/* If log stream was opened, close it now */
	if (logfp != NULL)
		(void) fclose(logfp);

	/* Close listening socket */
	if (lsn != -1)
		(void) close(lsn);
}

int
main(int argc, char *argv[])
{
	int		i;
	const char	*optarg;

	/*
	 * Parse command-line options. Not every system has getopt()
	 * function, so we use the good old method.
	 */
#define	GETARG(p)							\
	if ((p = (argv[i][2] ? &argv[i][2] : argv[++i])) == NULL)	\
		usage(argv[0])

	for (i = 1; i < argc && argv[i][0] == '-'; i++)
		switch (argv[i][1]) {

		case 'd':		/* Document root */
			GETARG(optarg);
			docroot = optarg;
			break;

		case 'p':		/* Listening port */
			GETARG(optarg);
			lport = (uint16_t) atoi(optarg);
			break;

		case 'l':		/* Log file */
			GETARG(optarg);
			logfile = optarg;
			break;

		case 'i':		/* Index file */
			GETARG(optarg);
			ifile = optarg;
			break;

		case 'P':		/* Passwords file */
			GETARG(optarg);
			globhtpass = optarg;
			break;

		case 'c':		/* CGI pattern */
			GETARG(optarg);
			cgipat = optarg;
			break;

		case 'S':		/* SSL mode */
			dossl++;
			lport = 443;
			break;

		case 's':		/* Certificate */
			GETARG(optarg);
			certfile = optarg;
			break;

		case 'u':		/* Runtime UID */
			GETARG(optarg);
			uid = atoi(optarg);
			break;
		
		case 'e':		/* Error file */
			GETARG(optarg);
			errfile = optarg;
			break;

		case 'D':		/* Directory listing */
			do_dir++;
			break;

		case 'N':		/* Server Name */
			GETARG(optarg);
			domain = optarg;
			break;

		case 'm':		/* Mime types file */
			GETARG(optarg);
			mimefile = optarg;
			break;

		default:
			usage(argv[0]);
			break;
		}

	doit();

	return (EXIT_SUCCESS);
}
